package xCAT_plugin::ibswitch;
use Data::Dumper;
use Sys::Syslog;
use Socket;
use xCAT::MsgUtils;

sub handled_commands {
    return {
	findme => 'ibswitch',
    };
}


my $utildata;

#
# exactly the same function as xCAT::Utils->nodeonmynet(), but on
# success returns network interface name (not just 1)
#
sub node_on_my_which_net
{
    my $nodetocheck = shift;
    if (scalar(@_))
    {
        $nodetocheck = shift;
    }
    unless (inet_aton($nodetocheck))
    {
        return 0;
    }
    my $nodeip = inet_ntoa(inet_aton($nodetocheck));
    unless ($nodeip =~ /\d+\.\d+\.\d+\.\d+/)
    {
        return 0;    #Not supporting IPv6 here IPV6TODO
    }
    my $noden = unpack("N", inet_aton($nodeip));
    my @nets;
    if ($utildata->{nodeonmynetdata} and $utildata->{nodeonmynetdata}->{pid} == $$) {
        @nets = @{$utildata->{nodeonmynetdata}->{nets}};
    } else {
        if ( $^O eq 'aix')
        {
            my @subnets = get_subnet_aix();
            for my $net_ent (@subnets)
            {
                my @ents = split /:/ , $net_ent;
                push @nets, $ents[0] . '/' . $ents[2] . ' dev ' . $ents[1];
            }
	    
        }
        else
        {
            @nets = split /\n/, `/sbin/ip route`;
        }
        my $nettab=xCAT::Table->new("networks");
        my @vnets = $nettab->getAllAttribs('net','mgtifname','mask');
        foreach (@vnets) {
            if ($_->{mgtifname} eq '!remote!') { #global scoped network
                my $curm = unpack("N", inet_aton($_->{mask}));
                my $bits=32;
                until ($curm & 1)  {
                    $bits--;
                    $curm=$curm>>1;
                }
                push @nets,$_->{'net'}."/".$bits." dev remote";
            }
        }
        $utildata->{nodeonmynetdata}->{pid}=$$;
        $utildata->{nodeonmynetdata}->{nets} = \@nets;
    }
    foreach (@nets)
    {
        my @elems = split /\s+/;
        unless ($elems[1] =~ /dev/)
        {
            next;
        }
        (my $curnet, my $maskbits) = split /\//, $elems[0];
        my $curmask = 2**$maskbits - 1 << (32 - $maskbits);
        my $curn = unpack("N", inet_aton($curnet));
        if (($noden & $curmask) == $curn)
        {
#            return 1;
            return $elems[2];
        }
    }
    return 0;
}

sub mac2gid {
    
    my $mac = shift;
    if (scalar(split(/:/, $mac)) >= 8) {
	my ($gid) = $mac =~ /^.*((:*[0-9a-fA-F]{2}){8})$/;
	$gid =~ s/://g;
	return "0x$gid";
    }
    return 0;
}    

sub get_ib_switch_and_port {
    
    my $src_gid = shift;
    my $dst_gid = shift;
    my $l;
    
    my @trace = split /\n/, `ibtracert -n -G $src_gid $dst_gid`;
    
# parsing rely on exact output format, example:
#
#> ibtracert -n -G 0x003048c494c00001 0x003048c4e3d60001
#From {0x003048c494c00000}[1]
#[1] -> {0x000b8cffff005679}[8]
#[6] -> {0x003048c4e3d60001}[1]
#To {0x003048c4e3d60000}[1]
    
    for ($l=0; @trace[$l]; $l++) {
	if (@trace[$l] =~ /{$dst_gid}/) {
	    my ($switch_port) = @trace[$l] =~ /^\[(\d+)\] -> {$dst_gid}\[\d+\]$/; 
	    my ($switch_gid) = @trace[$l-1] =~ /^\[\d+\] -> {(.+)}\[\d+\]$/; 
	    return ($switch_gid, $switch_port);
	}
    }
    return undef;
}

sub get_node {
    
    my ($sw_gid, $sw_port) = @_;
    my $ibswitchtab = xCAT::Table->new('ibswitch', -create => 1);
    unless ($ibswitchtab) {
	xCAT::MsgUtils->message("S","ERROR: Unable to open ibswitch table");
	return undef;
    }
    
    my @entries = $ibswitchtab->getAllNodeAttribs(['node','switch', 'port']);
    $ibswitchtab->close;

    for my $entry (@entries) {
        if (defined($entry->{switch}) and defined($entry->{port}) and
            $entry->{switch} eq $sw_gid and $entry->{port} == $sw_port) {
            return $entry->{node};
        }
    }
    xCAT::MsgUtils->message("S","No node defined for switch: $sw_gid, port: $sw_port in ibswitch table, try: \nchtab ibswitch.switch=$sw_gid ibswitch.port=$sw_port ibswitch.node=NODE_NAME");
    return undef;
}


sub process_request {

    my $req = shift;
    my $cb = shift;
    my $doreq = shift;
    my $ip = $req->{'_xcat_clientip'};
    my $mac = '';
    my $iface;
    my $node;
    my $firstpass=1;
    my $cmd=$req->{command}->[0];
    
    if (not $node) {
	foreach (@{$req->{mac}}) {

	    (my $drv, $iface, $mac, $ip) = $_ =~ /(.*)\|(.*)\|([\dABCDEFabcdef:]+)(\|.*|)$/;
	    $ip =~ s/\|//g;
	    
	    next unless $iface =~ /^ib/ && $ip && $mac;

	    my $mn_iface = node_on_my_which_net((split /\//, $ip)[0]);
	    next unless $mn_iface;

	    my $gid = mac2gid ($mac);
	    next unless $gid;
	    
	    my $mn_gid = mac2gid (`cat /sys/class/net/$mn_iface/address`);
	    next unless $mn_gid;
	    
	    my ($sw_gid, $sw_port) = get_ib_switch_and_port ($mn_gid, $gid);
	    next unless ($sw_gid && $sw_port);
	    
	    $node = get_node ($sw_gid, $sw_port);
	    
	    $firstpass=0;
	    if ($node) { last; }
	}
    }
    if ($node) {

	my $mactab = xCAT::Table->new('mac',-create=>1);
	$mactab->setNodeAttribs($node,{mac=>$mac, interface=>$iface});
#	xCAT::MsgUtils->message("S","::ibswitch:: add node: $node, mac: $mac, interface: $iface");
	$mactab->close();
	$req->{command}=['discovered'];
	$req->{noderange} = [$node];
	$doreq->($req); 
	%{$req}=();#Clear req structure, it's done..
	undef $mactab;
    }
}
1;

